from __future__ import absolute_import

import datetime
import decimal
import itertools
import json

import six
import sqlalchemy as sa
from six.moves import zip
from sqlalchemy import TypeDecorator, UnicodeText
from sqlalchemy.engine.reflection import Inspector


def result_to_dict(res):
    """
    :param res: :any:`sqlalchemy.engine.ResultProxy`

    :return: a list of dicts where each dict represents a row in the query where the key \
    is the column name and the value is the value of that column.
    """
    keys = res.keys()
    all_ = res.fetchall()
    return [dict(zip(keys, row)) for row in all_]


def generate_where_clause(cls, row, col, use_dirty=True):
    """
    :param cls: the sqlalchemy ORM model
    :param row: a sqlalchemy ORM model object (must be an instance of :py:data:`cls`)
    :param col: the column name
    :param use_dirty: if ``True`` (default) will return the dirty value of the column

    :return: a sqlalchemy ``==`` clause
    """
    return getattr(cls, col) == get_column_attribute(row, col, use_dirty=use_dirty)


def generate_and_clause(cls, row, cols, use_dirty=True):
    """
    :param cls: the sqlalchemy ORM model
    :param row: a sqlalchemy ORM model object (must be an instance of :py:data:`cls`)
    :param cols: an iterable of strings corresponding to column names
    :param use_dirty: if ``True`` (default) will return the dirty value of the column

    :return: a :py:func:`sqlalchemy.and_` clause which checks for equality of all columns \
    in cols to the value they contain in row.

    For example,

    .. code-block:: python

        generate_and_clause(cls, ['foo', 'bar'], cols) =
    would return

    .. code-block:: python

        sqlalchemy.and_(cls.foo == row.foo, cls.bar == row.bar)
    """

    return sa.and_(*(
        generate_where_clause(cls, row, col_name, use_dirty=use_dirty)
        for col_name in cols
    ))


def get_bind_processor(row, col_name, dialect):
    '''
    Returns a bind_processor for the given column in the row based on the dialect. If dialect
    is None or there is no bind_processor, returns the identity function. The return value
    of this can be applied to ``getattr(row, col_name)`` and will return the sql type of that value.
    '''
    bind_processor = None
    if dialect is not None:
        bind_processor = getattr(type(row), col_name).type.bind_processor(dialect)
    return (lambda x: x) if bind_processor is None else bind_processor


def get_column_attribute(row, col_name, use_dirty=True, dialect=None):
    """
    :param row: the row object
    :param col_name: the column name
    :param use_dirty: whether to return the dirty value of the column
    :param dialect: if not None, should be a :py:class:`~sqlalchemy.engine.interfaces.Dialect`. If \
        specified, this function will process the column attribute into the dialect type before \
        returning it; useful if one is using user defined column types in their mappers.

    :return: if :any:`use_dirty`, this will return the value of col_name on the row before it was \
    changed; else this will return getattr(row, col_name)
    """
    bind_processor = get_bind_processor(row, col_name, dialect)
    hist = getattr(sa.inspect(row).attrs, col_name).history
    getattr(row, col_name)
    if not use_dirty and hist.has_changes():
        if hist.deleted:
            return bind_processor(hist.deleted[0])
        else:
            return None
    return bind_processor(getattr(row, col_name))


def get_column_keys(table):
    '''
    Return a generator of names of the python attribute for the table columns.
    '''
    for k, _ in get_column_keys_and_names(table):
        yield k


def get_column_names(table):
    '''
    Return a generator of names of the name of the column in the sql table.
    '''
    for _, c in get_column_keys_and_names(table):
        yield c


def get_column_keys_and_names(table):
    '''
    Return a generator of tuples k, c such that k is the name of the python attribute for
    the column and c is the name of the column in the sql table.
    '''
    ins = sa.inspect(table)
    return ((k, c.name) for k, c in ins.mapper.c.items())


def get_dialect(session):
    return session.connection().dialect


def has_constraint(tbl_name, engine, *col_names):
    """
    :param tbl_name: a string with the name of the table to check
    :param engine: an instance of :class:`sa.engine.Engine` from which to execute the query
    :param col_names: the name of columns which the unique constraint should contain

    :rtype: bool
    :return: True if the given columns are part of a unique constraint on tbl_name
    """
    insp = Inspector.from_engine(engine)
    constraints = itertools.chain(
        (sorted(x['column_names']) for x in insp.get_unique_constraints(tbl_name)),
        sorted(insp.get_pk_constraint(tbl_name)['constrained_columns']),
    )
    return sorted(col_names) in constraints


def is_modified(row, ignore=None):
    if ignore is None:
        ignore = set()
    ins = sa.inspect(row)
    modified_cols = \
        {key for key in get_column_keys(ins.mapper) if key not in ignore} - ins.unmodified
    return not all((
        get_column_attribute(row, col) == get_column_attribute(row, col, use_dirty=False)
        for col in modified_cols
    ))


class VAJSONEncoder(json.JSONEncoder):
    """
    Extends the default encoder to add support for serializing datetime objects and decimals.

    Datetimes are serialized using `datetime.isoformat()`; the resulting string can be reloaded
    into a MySQL/Postgres TIMESTAMP column directly (verified on MySQL 5.6 and Postgres 9.6).
    Decimals are serialized using `float(Decimal)`.
    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super(VAJSONEncoder, self).default(obj)


class _JSONEncoded(TypeDecorator):
    """
    Does validation and serde on a JSON python type (list, dict, int, str) to
    a text based column in a SQL database. This class should be overriden for each
    primitive JSON type.
    """

    impl = UnicodeText
    json_type = None

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        elif isinstance(value, six.string_types):
            value = json.loads(value)

        if self.json_type is not None and not isinstance(value, self.json_type):
            raise ValueError('value of type {} is not {}'.format(type(value), self.json_type))

        return json.dumps(value, ensure_ascii=False, cls=VAJSONEncoder)

    def process_result_value(self, value, dialect):
        if value is None:
            return None

        value = json.loads(value)
        if self.json_type is not None and not isinstance(value, self.json_type):
            raise ValueError('value of type {} is not {}'.format(type(value), self.json_type))

        return value


class JSONEncodedList(_JSONEncoded):
    json_type = list


class JSONEncodedDict(_JSONEncoded):
    json_type = dict



from __future__ import absolute_import

import six
import sqlalchemy as sa

from versionalchemy import utils


def delete(va_table, session, conds):
    '''
    :param va_table: the model class which inherits from \
        :class:`~versionalchemy.models.user_table.VAModelMixin` and specifies the model \
        of the user table from which we are querying
    :param session: a sqlalchemy session with connections to the database
    :param conds: a list of dictionary of key value pairs where keys are columns in the table \
        and values are values the column should take on. If specified, this query will \
        only return rows where the columns meet all the conditions. The columns specified \
        in this dictionary must be exactly the unique columns that versioning pivots around.

    Performs a hard delete on a row, which means the row is deleted from the versionalchemy \
    table as well as the archive table.
    '''
    with session.begin_nested():
        archive_conds_list = _get_conditions_list(va_table, conds)
        session.execute(
            sa.delete(va_table.ArchiveTable, whereclause=_get_conditions(archive_conds_list))
        )
        conds_list = _get_conditions_list(va_table, conds, archive=False)
        session.execute(
            sa.delete(va_table, whereclause=_get_conditions(conds_list))
        )


def get(
    va_table,
    session,
    va_id=None,
    t1=None,
    t2=None,
    fields=None,
    conds=None,
    include_deleted=True,
    page=1,
    page_size=100,
):
    '''
    :param va_table: the model class which inherits from \
        :class:`~versionalchemy.models.user_table.VAModelMixin` and specifies the model of \
        the user table from which we are querying
    :param session: a sqlalchemy session with connections to the database
    :param va_id: if specified, the value of t1 and t2 will be ignored. If specified, this will \
        return all records after the specified va_id.
    :param t1: lower bound time for this query; if None or unspecified, \
        defaults to the unix epoch. If this is specified and t2 is not, this query \
        will simply return the time slice of data at t1. This must either be a valid \
        sql time string or a datetime.datetime object.
    :param t2: upper bound time for this query; if both t1 and t2 are none or unspecified, \
        this will return the latest data (i.e. time slice of data now). This must either be a \
        valid sql time string or a datetime.datetime object.
    :param fields: a list of strings which corresponds to columns in the table; If \
        None or unspecified, returns all fields in the table.
    :param conds: a list of dictionary of key value pairs where keys are columns in the table \
        and values are values the column should take on. If specified, this query will \
        only return rows where the columns meet all the conditions. The columns specified \
        in this dictionary must be exactly the unique columns that versioning pivots around.
    :param include_deleted: if ``True``, the response will include deleted changes. Else it will \
        only include changes where ``deleted = 0`` i.e. the data was in the user table.
    :param page: the offset of the result set (1-indexed); i.e. if page_size is 100 and page is 2, \
        the result set will contain results 100 - 199
    :param page_size: upper bound on number of results to display. Note the actual returned result \
        set may be smaller than this due to the roll up.
    '''
    limit, offset = _get_limit_and_offset(page, page_size)
    version_col_names = va_table.va_version_columns
    if fields is None:
        fields = [name for name in utils.get_column_names(va_table) if name != 'va_id']

    if va_id is not None:
        return _format_response(utils.result_to_dict(session.execute(
            sa.select([va_table.ArchiveTable])
            .where(va_table.ArchiveTable.va_id > va_id)
            .order_by(*_get_order_clause(va_table.ArchiveTable))
            .limit(page_size)
            .offset(offset)
        )), fields, version_col_names)

    if t1 is None and t2 is None:
        rows = _get_latest_time_slice(va_table, session, conds, include_deleted, limit, offset)
        return _format_response(rows, fields, version_col_names)

    if t2 is None:  # return a historical time slice
        rows = _get_historical_time_slice(
            va_table, session, t1, conds, include_deleted, limit, offset
        )
        return _format_response(rows, fields, version_col_names)

    if t1 is None:
        t1 = 0

    rows = _get_historical_changes(
        va_table, session, conds, t1, t2, include_deleted, limit, offset
    )
    return _format_response(rows, fields, version_col_names)


def _format_response(rows, fields, unique_col_names):
    '''
    :param rows: a list of dictionaries representing rows from the ArchiveTable.
    :param fields: a list of strings of fields to be extracted from the archived row.

    This function will look at the data column of rows and extract the specified fields. It
    will also dedup changes where the specified fields have not changed. The list of rows should
    be ordered by the compound primary key which versioning pivots around and be in ascending
    version order.

    This function will return a list of dictionaries where each dictionary has the following
    schema:
        {
            'updated_at': timestamp of the change,
            'version': version number for the change,
            'data': a nested dictionary containing all keys specified in fields and values
                corresponding to values in the user table.
        }

    Note that some versions may be omitted in the output for the same key if the specified fields
    were not changed between versions.
    '''
    output = []
    old_id = None
    for row in rows:
        id_ = {k: row[k] for k in unique_col_names}
        formatted = {k: row[k] for k in row if k != 'va_data'}
        if id_ != old_id:  # new unique versioned row
            data = row['va_data']
            formatted['va_data'] = {k: data.get(k) for k in fields}
            output.append(formatted)
        else:
            data = row['va_data']
            pruned_data = {k: data.get(k) for k in fields}
            if (
                pruned_data != output[-1]['va_data'] or
                row['va_deleted'] != output[-1]['va_deleted']
            ):
                formatted['va_data'] = pruned_data
                output.append(formatted)
        old_id = id_
    return output


def _get_conditions(pk_conds, and_conds=None):
    '''
    :param pk_conds: a list of list of primary key constraints returned by _get_conditions_list
    :param and_conds: additional and conditions to be placed on the query

    If and_conds = [a1, a2, ..., an] and pk_conds = [[b11, b12, ..., b1m], ... [bk1, ..., bkm]],
    this function will return the mysql condition clause:
        a1 & a2 & ... an & ((b11 and ... b1m) or ... (b11 and ... b1m))
    '''
    if and_conds is None:
        and_conds = []

    if len(and_conds) == 0 and len(pk_conds) == 0:
        return sa.and_()

    condition1 = sa.and_(*and_conds)
    condition2 = sa.or_(*[sa.and_(*cond) for cond in pk_conds])
    return sa.and_(condition1, condition2)


def _get_conditions_list(va_table, conds, archive=True):
    '''
    :param va_table: the user table model class which inherits from \
        versionalchemy.models.VAModelMixin
    :param conds: a list of dictionaries of key value pairs where keys are column names and \
        values are conditions to be placed on the column.
    :param archive: If true, the condition is with columns from the archive table. Else its from \
        the user table.

    This function returns a list of list of == conditions on sqlalchemy columns given conds. \
    This should be treated as an or of ands.
    '''
    if conds is None:
        conds = []

    all_conditions = []
    for cond in conds:
        if len(cond) != len(va_table.va_version_columns):
            raise ValueError('Conditions must specify all unique constraints.')

        conditions = []
        t = va_table.ArchiveTable if archive else va_table

        for col_name, value in six.iteritems(cond):
            if col_name not in va_table.va_version_columns:
                raise ValueError('{} is not one of the unique columns <{}>'.format(
                    col_name, ','.join(va_table.va_version_columns)
                ))
            conditions.append(getattr(t, col_name) == value)
        all_conditions.append(conditions)
    return all_conditions


def _get_historical_changes(va_table, session, conds, t1, t2, include_deleted, limit, offset):
    pk_conditions = _get_conditions_list(va_table, conds)
    and_clause = _get_conditions(
        pk_conditions,
        [va_table.ArchiveTable.va_updated_at >= t1, va_table.ArchiveTable.va_updated_at < t2] +
        [] if include_deleted else [va_table.ArchiveTable.va_deleted.is_(False)],
    )

    return utils.result_to_dict(session.execute(
        sa.select([va_table.ArchiveTable])
        .where(and_clause)
        .order_by(*_get_order_clause(va_table.ArchiveTable))
        .limit(limit)
        .offset(offset)
    ))


def _get_historical_time_slice(va_table, session, t, conds, include_deleted, limit, offset):
    at = va_table.ArchiveTable
    vc = va_table.va_version_columns
    pk_conditions = _get_conditions_list(va_table, conds)
    and_clause = _get_conditions(
        pk_conditions,
        [at.va_updated_at <= t] +
        [] if include_deleted else [va_table.ArchiveTable.va_deleted.is_(False)],
    )
    t2 = at.__table__.alias('t2')
    return utils.result_to_dict(session.execute(
        sa.select([at])
        .select_from(at.__table__.join(
            t2,
            sa.and_(
                t2.c.va_updated_at <= t,
                at.va_version < t2.c.va_version,
                *[getattr(at, c) == getattr(t2.c, c) for c in vc]
            ),
            isouter=True,
        ))
        .where(t2.c.va_version.is_(None) & and_clause)
        .order_by(*_get_order_clause(at))
        .limit(limit)
        .offset(offset)
    ))


def _get_latest_time_slice(va_table, session, conds, include_deleted, limit, offset):
    and_clause = _get_conditions(
        _get_conditions_list(va_table, conds, archive=False),
        [] if include_deleted else [va_table.ArchiveTable.va_deleted.is_(False)],
    )
    result = session.execute(
        sa.select([va_table.ArchiveTable]).select_from(
            va_table.ArchiveTable.__table__.join(
                va_table,
                va_table.ArchiveTable.va_id == va_table.va_id
            )
        )
        .where(and_clause)
        .order_by(*_get_order_clause(va_table.ArchiveTable))
        .limit(limit)
        .offset(offset)
    )
    return utils.result_to_dict(result)


def _get_limit_and_offset(page, page_size):
    '''
    Returns a 0-indexed offset and limit based on page and page_size for a MySQL query.
    '''
    if page < 1:
        raise ValueError('page must be >= 1')
    limit = page_size
    offset = (page - 1) * page_size
    return limit, offset


def _get_order_clause(archive_table):
    '''
    Returns an ascending order clause on the versioned unique constraint as well as the
    version column.
    '''
    order_clause = [
        sa.asc(getattr(archive_table, col_name)) for col_name in archive_table._version_col_names
    ]
    order_clause.append(sa.asc(archive_table.va_version))
    return order_clause



from __future__ import absolute_import

from setuptools import setup

install_requires = [
    "six>=1.12.0",
    "SQLAlchemy>=1.0",
]
dev_requires = [
    "autopep8>=1.4.4",
    "coverage",
    "flake8",
    "ipdb",
    "ipython",
    "isort>=4.3.21",
    'mock;python_version=="2.7"',
    "nose",
    "pip-tools",
    "pylint",
    "sphinx",
    "sphinx-rtd-theme<0.2",
]

with open("VERSION") as version_fd:
    version = version_fd.read().strip()
with open("README.rst") as f:
    long_description = f.read()
url = "https://github.com/NerdWalletOSS/versionalchemy"
download_url = "{}/tarball/v{}".format(url, version)
classifiers = """
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: SQL
Topic :: Database
Topic :: Database :: Front-Ends
Topic :: Software Development
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: OS Independent
"""

setup(
    name='versionalchemy',
    version=version,
    author='Akshay Nanavati',
    author_email='akshay@nerdwallet.com',
    maintainer="Jeremy Lewis",
    maintainer_email="jlewis@nerdwallet.com",
    url=url,
    description='Versioning library for relational data',
    long_description=long_description,
    download_url=download_url,
    classifiers=[c for c in classifiers.split("\n") if c],
    license='MIT License',
    packages=['versionalchemy', 'versionalchemy/api', 'versionalchemy/models'],
    install_requires=install_requires,
    extras_require={"dev": dev_requires},
    # Currently `versionalchemy` supports Python 2.7, and Python 3.6+
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4",
    include_package_data=True,
)



from __future__ import absolute_import

import sqlalchemy as sa
from sqlalchemy.orm import Session

from versionalchemy import utils
from versionalchemy.exceptions import LogTableCreationError

from .models import VAModelMixin

_initialized = False


def init():
    """
    :param Session: the session factory class
    """
    global _initialized

    if _initialized:
        return
    _initialized = True
    sa.event.listen(Session, 'after_flush', _after_flush_handler)


def is_initialized():
    global _initialized
    return _initialized


def _after_flush_handler(session, flush_context):
    handlers = [
        (_versioned_delete, session.deleted),
        (_versioned_insert, session.new),
        (_versioned_update, session.dirty),
    ]
    for handler, rows in handlers:
        for row in rows:
            if isinstance(row, VAModelMixin):
                if not hasattr(row, 'ArchiveTable'):
                    raise LogTableCreationError('Need to register va tables!!')
                user_id = getattr(row, '_updated_by', None)
                va_id = handler(row, session, user_id)
                if va_id:
                    Model = type(row)
                    where_clause = utils.generate_and_clause(
                        Model, row, row.va_version_columns
                    )
                    session.execute(sa.update(Model, values={'va_id': va_id}).where(where_clause))
                    row.va_id = va_id


def _versioned_delete(row, session, user_id=None):
    result = session.execute(
        sa.insert(row.ArchiveTable),
        row.ArchiveTable.build_row_dict(row, session, deleted=True, user_id=user_id)
    )
    # This should not matter since the row is being deleted anyways from the user
    # table but is here if we ever decide to implement soft deletes on the user table
    return result.inserted_primary_key[0]


def _versioned_update(row, session, user_id=None):
    if not utils.is_modified(row, ignore={'va_id'}):
        return

    # Check if composite key has been changed
    for col in row.va_version_columns:
        hist = getattr(sa.inspect(row).attrs, col).history
        if hist.has_changes():
            # delete the original row from the archive table
            session.execute(
                sa.insert(row.ArchiveTable),
                row.ArchiveTable.build_row_dict(
                    row, session, user_id=user_id, deleted=True, use_dirty=False
                )
            )

    result = session.execute(
        sa.insert(row.ArchiveTable),
        row.ArchiveTable.build_row_dict(row, session, user_id=user_id)
    )
    return result.inserted_primary_key[0]


def _versioned_insert(row, session, user_id=None):
    result = session.execute(
        sa.insert(row.ArchiveTable),
        row.ArchiveTable.build_row_dict(row, session, user_id=user_id)
    )
    return result.inserted_primary_key[0]



from __future__ import absolute_import

from datetime import datetime

import sqlalchemy as sa
from sqlalchemy import Boolean, Column, DateTime, func, Integer

from versionalchemy import utils
from versionalchemy.exceptions import LogTableCreationError


class VALogMixin(object):
    """
    A mixin providing the schema for the log table, an append only table which saves old versions
    of rows. An inheriting model must specify the following columns:
      - user_id - a column corresponding to the user that made the specified change
      - 1 or more columns which are a subset of columns in the user table. These columns
      must have a unique constraint on the user table and also be named the same in both tables
    """
    va_id = Column(Integer, primary_key=True, autoincrement=True)
    va_version = Column(Integer, nullable=False, index=True)
    va_deleted = Column(Boolean, nullable=False)
    va_updated_at = Column(DateTime, nullable=False)
    va_data = Column(utils.JSONEncodedDict, nullable=False)  # JSON blob

    @classmethod
    def build_row_dict(cls, ut_row, session, deleted=False, user_id=None, use_dirty=True):
        """
        :param ut_row: the row from the user table
        :param deleted: whether or not the row is deleted
        :param user_id: the user that is performing the update on this row
        :param use_dirty: whether to use the dirty fields from ut_row or not

        :return: a dictionary of key value pairs to be inserted into the archive table
        :rtype: dict
        """
        at_data = {
            'va_deleted': deleted,
            'va_updated_at': datetime.now(),
            'va_data': ut_row._to_dict(utils.get_dialect(session), use_dirty=use_dirty),
        }
        version = cls._latest_version(session, ut_row, use_dirty=use_dirty)
        at_data['va_version'] = 0 if version is None else version + 1
        for col_name in cls._version_col_names:
            at_data[col_name] = utils.get_column_attribute(ut_row, col_name, use_dirty=use_dirty)

        if user_id is not None:
            at_data['user_id'] = user_id

        return at_data

    @classmethod
    def _latest_version(cls, session, row, use_dirty=True):
        """
        :param session: a session instance to execute a select on the log table
        :param row: an instance of the user table row object

        :return: the maximum version ID recorded for the specified row or None if version_id
        has not been inserted
        :rtype: int
        """
        and_clause = \
            utils.generate_and_clause(cls, row, cls._version_col_names, use_dirty=use_dirty)
        result = session.execute(
            sa.select([func.max(cls.va_version)]).
            where(and_clause)
        ).first()
        return None if result is None else result[0]

    @classmethod
    def _validate(cls, engine, *version_cols):
        """
        :param engine: instance of :class:`~sa.engine.Engine`
        :param *version_cols: instances of :class:`~sa.orm.attributes.InstrumentedAttribute` from
        the user table corresponding to the columns that versioning pivots around

        If all of the properties are not met, this function raises :class:`~LogTableCreationError`:
            - all version columns exist in the archive table
            - the python types of the user table and archive table columns are the same
            - a user_id column exists
            - there is a unique constraint on version and the other versioned columns from the
            user table
        """
        cls._version_col_names = set()
        for version_column_ut in version_cols:
            # Make sure all version columns exist on this table
            version_col_name = version_column_ut.key
            version_column_at = getattr(cls, version_col_name, None)
            if not isinstance(version_column_at, sa.orm.attributes.InstrumentedAttribute):
                raise LogTableCreationError(
                    "Log table needs {} column".format(version_col_name)
                )

            # Make sure the type of the user table and log table columns are the same
            version_col_at_t = version_column_at.property.columns[0].type.python_type
            version_col_ut_t = version_column_ut.property.columns[0].type.python_type
            if version_col_at_t != version_col_ut_t:
                raise LogTableCreationError(
                    "Type of column {} must match in log and user table".format(version_col_name)
                )
            cls._version_col_names.add(version_col_name)

        # Ensure user added a user_id column
        # TODO: should user_id column be optional?
        user_id = getattr(cls, 'user_id', None)
        if not isinstance(user_id, sa.orm.attributes.InstrumentedAttribute):
            raise LogTableCreationError(
                "Log table needs user_id column"
            )

        # Check the unique constraint on the versioned columns
        version_col_names = list(cls._version_col_names) + ['va_version']
        if not utils.has_constraint(cls.__tablename__, engine, *version_col_names):
            raise LogTableCreationError(
                "There is no unique contraint on the version columns"
            )


class VAModelMixin(object):
    va_id = Column(Integer, nullable=False, default=0)

    va_ignore_columns = None
    va_version_columns = None

    def updated_by(self, user):
        self._updated_by = user

    @classmethod
    def register(cls, ArchiveTable, engine):
        """
        :param ArchiveTable: the model for the users archive table
        :param engine: the database engine
        :param version_col_names: strings which correspond to columns that versioning will pivot \
            around. These columns must have a unique constraint set on them.
        """
        version_col_names = cls.va_version_columns
        if not version_col_names:
            raise LogTableCreationError('Need to specify version cols in cls.va_version_columns')
        if cls.va_ignore_columns is None:
            cls.va_ignore_columns = set()
        cls.va_ignore_columns.add('va_id')
        version_cols = [getattr(cls, col_name, None) for col_name in version_col_names]

        cls._validate(engine, *version_cols)

        ArchiveTable._validate(engine, *version_cols)
        cls.ArchiveTable = ArchiveTable

    def _to_dict(self, dialect, use_dirty=True):
        """
        :param dialect: a :py:class:`~sqlalchemy.engine.interfaces.Dialect` corresponding to the \
            SQL dialect being used.
        :param use_dirty: whether to make a dict of the fields as they stand, or the fields \
            before the row was updated

        :return: a dictionary of key value pairs representing this row.
        :rtype: dict
        """
        return {
            cn: utils.get_column_attribute(self, c, use_dirty=use_dirty, dialect=dialect)
            for c, cn in utils.get_column_keys_and_names(self)
            if c not in self.va_ignore_columns
        }

    @classmethod
    def _validate(cls, engine, *version_cols):
        version_col_names = set()
        for version_column_ut in version_cols:
            if not isinstance(version_column_ut, sa.orm.attributes.InstrumentedAttribute):
                raise LogTableCreationError(
                    "All version columns must be <sa.orm.attributes.InstrumentedAttribute>"
                )
            version_col_names.add(version_column_ut.key)

        # Check the unique constraint on the versioned columns
        insp = sa.inspect(cls)
        uc = sorted([col.name for col in insp.primary_key]) == sorted(version_col_names)
        if not (uc or utils.has_constraint(cls.__tablename__, engine, *version_col_names)):
            raise LogTableCreationError(
                "There is no unique contraint on the version columns"
            )

    def version(self, session):
        """
        Returns the rows current version. This can only be called after a row has been
        inserted into the table and the session has been flushed. Otherwise this
        method has undefined behavior.
        """
        result = session.execute(
            sa.select([self.ArchiveTable.va_version]).
            where(self.ArchiveTable.va_id == self.va_id)
        ).first()
        return result[0]
